# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Because this test may indirectly import pylab (e.g. via importing
# bdbcontrib.crosscat_utils), which would peg the matplotlib backend,
# which might prevent a later test from drawing pictures headless.
# &#&*%(@#^&!@.
import matplotlib
matplotlib.use('Agg')

import bayeslite
from bayeslite.exception import BayesLiteException as BLE
import pandas
import pytest

from bayeslite.read_pandas import bayesdb_read_pandas_df
from bdbcontrib import crosscat_utils

def get_test_df():
    PANDAS_DF_DATA = [
        {
            'age': 34,
            'gender': 'M',
            'salary': 7400,
            'height': 65,
            'division': 'sales',
            'rank': 3
            },
        {
            'age': 41,
            'gender': 'M',
            'salary': 6560,
            'height': 72,
            'division': 'marketing',
            'rank': 4
            },
        {
            'age': 25,
            'gender': 'M',
            'salary': 5200,
            'height': 69,
            'division':
            'accounting',
            'rank': 5
            },
        {
            'age': 23,
            'gender': 'F',
            'salary': 8100,
            'height': 67,
            'division':
            'data science',
            'rank': 3
            },
        {
            'age': 36,
            'gender': 'F',
            'salary': 9600,
            'height': 70,
            'division': 'management',
            'rank': 2
            },
        {
            'age': 30,
            'gender': 'M',
            'salary': 7000,
            'height': 73,
            'division': 'sales',
            'rank': 4
            },
        {
            'age': 30,
            'gender': 'F',
            'salary': 8100,
            'height': 73,
            'division': 'engineering',
            'rank': 3
            },
    ]
    return pandas.DataFrame(PANDAS_DF_DATA)


def test_get_metadata():
    table_name = 'tmp_table'
    generator_name = 'tmp_cc'
    pandas_df = get_test_df()

    import os
    os.environ['BAYESDB_WIZARD_MODE']='1'
    with bayeslite.bayesdb_open() as bdb:
        bayesdb_read_pandas_df(bdb, table_name, pandas_df, create=True)
        bdb.execute('''
            create generator {} for {} using crosscat(guess(*))
        '''.format(generator_name, table_name))
        with pytest.raises(BLE):
            md = crosscat_utils.get_metadata(bdb, generator_name, 0)

        bdb.execute('INITIALIZE 2 MODELS FOR {}'.format(generator_name))

        with pytest.raises(ValueError):  # XXX from BayesLite: should be a BLE?
            crosscat_utils.get_metadata(bdb, 'Peter_Gabriel', 0)
        md = crosscat_utils.get_metadata(bdb, generator_name, 0)

        assert isinstance(md, dict)
        assert 'X_D' in md.keys()
        assert 'X_L' in md.keys()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Aggregation of probe results over bdb files and model numbers.

A single probe is a typed result of some computation over a BayesDB,
whose variation with model count, analysis iteration, or other setup
one is interested in studying.

A probe function is a function that accepts a bdb and computes a list
of probe results (more than one so that they can share intermediate
computations).

A probe result is represented as a 3-tuple: the name of the probe, the
type of the result (as a string), and the value of the result.

Two result types are currently supported:

- 'num' results are aggregated by maintaining a list of all of them.
- 'bool' results are aggregated by maintaining the count of True and
  False results seen.

"""

import contextlib
import logging
import random
import time

import bayeslite
from bayeslite.core import bayesdb_get_generator
from bayeslite.metamodels.crosscat import CrosscatMetamodel
from crosscat.LocalEngine import LocalEngine as CrosscatLocalEngine

start_time = time.time()
def log(msg, *irritants):
    logging.info("At %3.2fs " % (time.time() - start_time) + msg, *irritants)

def probe_fileset(files, generator, probes, model_schedule=None,
                    n_replications=None, seed=0):
    """Aggregate all the probes over all the given bdb files.

    Measure seed and model count variation within each file by varying
    model sets to which probes are applied.

    `files` is a list of filenames of saved BayesDBs to aggregate over.

    `generator` is a string naming the generator the probes read.

    `probes` is a list of probe functions.

    `model_schedule` is a list of numbers of models to probe.

    `n_replications` is the number of replications to do.

    `seed` is the initial entropy for any randomness embedded in
    BayesDB queries performed by the probe functions.  Default: 0.

    Each saved bdb must have the named generator, and that generator
    must have at least as many models as `n_replications *
    max(model_schedule)`.

    Probe functions are expected to utilize all available models to
    compute their results; application to different sets of models is
    accomplished by temporarily dropping unwanted models.

    Returns a dict mapping conditions to aggregated probe results.
    Each condition is a 3-tuple: Name of bdb file, number of models
    probed, name of probe.  The results are aggregated over
    replications: different sets of the desired number of models
    within each file.

    """
    # TODO Expect the number of models in the file to be at least
    # model_skip * n_replications; excess models are wasted.
    # TODO Default the model schedule and n_replications based on some
    # heuristic like the square root of the number of models in the
    # first file.
    model_skip = max(model_schedule)
    specs = model_specs(model_schedule, model_skip, n_replications)
    return do_probe_fileset(files, generator, probes, specs, seed)

def do_probe_fileset(files, generator, probes, specs, seed):
    # Keys are (file_name, model_ct, name); values are aggregated results
    results = {}
    prng = random.Random(seed)
    for fname in files:
        log("processing file %s" % fname)
        with bayeslite.bayesdb_open(fname, builtin_metamodels=False) as bdb:
            bayeslite.bayesdb_register_metamodel(bdb, new_cc_metamodel(prng))
            res = [((fname, model_ct, name), ress)
                   for ((model_ct, name), ress)
                   in run_probes(bdb, generator, probes, specs).iteritems()]
            incorporate(results, res)
    return results

def new_cc_metamodel(prng):
    return CrosscatMetamodel(CrosscatLocalEngine(seed=prng.randint(0, 2**31)))

def run_probes(bdb, generator, probes, specs):
    results = {} # Keys are (model_ct, name); values are aggregated results
    ct = num_models_in_bdb(bdb, generator)
    for spec in specs:
        (low, high) = spec
        model_ct = high - low
        with model_restriction(bdb, generator, spec, ct):
            log("probing models %d-%d" % (low, high-1))
            for probeset in probes:
                pres = [((model_ct, name), inc_singleton(ptype, res))
                        for (name, ptype, res) in probeset(bdb)]
                incorporate(results, pres)
    return results

def num_models_in_bdb(bdb, generator):
    generator_id = bayesdb_get_generator(bdb, generator)
    q = '''SELECT COUNT(*), MAX(modelno)
        FROM bayesdb_generator_model WHERE generator_id = ?'''
    (ct, maximum) = bdb.execute(q, (generator_id,)).next()
    assert ct == maximum+1, "Expected models to be numbered sequentially"
    return ct

def model_specs(model_schedule, model_skip, n_replications):
    def spec_at(location, size):
        """Return a 0-indexed model range
        inclusive of lower bound and exclusive of upper bound."""
        return (location * model_skip, location * model_skip + size)
    return [spec_at(location, size)
            for location in range(n_replications)
            for size in model_schedule]

@contextlib.contextmanager
def model_restriction(bdb, gen_name, spec, model_count):
    (low, high) = spec
    assert model_count >= high, "Not enough models in bdb"
    with bdb.savepoint_rollback():
        if low > 0:
            bdb.execute('''DROP MODELS 0-%d FROM %s''' % (low-1, gen_name))
        if model_count > high:
            bdb.execute('''DROP MODELS %d-%d FROM %s'''
                        % (high, model_count-1, gen_name))
        yield

def incorporate(running, new):
    for (key, more) in new:
        if key in running:
            running[key] = merge_one(running[key], more)
        else:
            running[key] = more

# Type-dependent aggregation of intermediate values (e.g., counting
# yes and no for booleans vs accumulating lists of reals).
def inc_singleton(ptype, val):
    if ptype == 'num':
        return ('num', [val])
    elif ptype == 'bool':
        if val:
            return ('bool', (1, 0))
        else:
            return ('bool', (0, 1))
    else:
        raise Exception("Unknown singleton type %s for %s" % (ptype, val))

def merge_one(so_far, val):
    (tp1, item1) = so_far
    (tp2, item2) = val
    if tp1 == 'num' and tp2 == 'num':
        item1.extend(item2)
        return ('num', item1)
    elif tp1 == 'bool' and tp2 == 'bool':
        (t1, f1) = item1
        (t2, f2) = item2
        return ('bool', (t1+t2, f1+f2))
    else:
        raise Exception("Type mismatch in merge into %s of %s" % (so_far, val))



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pickle

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier

from bayeslite.exception import BayesLiteException as BLE
import bdbcontrib
from bdbcontrib.predictors import predictor
from bdbcontrib.predictors import sklearn_utils as utils

class RandomForest(predictor.IBayesDBForeignPredictor):
    """A Random Forest foreign predictor.

    The `targets` must be a single categorical stattype.  The `conditions`
    may be arbitrary numerical or categorical columns.
    """

    @classmethod
    def create(cls, bdb, table, targets, conditions):
        cols = [c for c,_ in targets+conditions]
        df = bdbcontrib.bql_utils.table_to_df(bdb, table, cols)
        rf = cls()
        rf.train(df, targets, conditions)
        rf.prng = bdb.np_prng
        return rf

    @classmethod
    def serialize(cls, _bdb, pred):
        state = {
            'targets': pred.targets,
            'conditions_numerical': pred.conditions_numerical,
            'conditions_categorical': pred.conditions_categorical,
            'rf_full': pred.rf_full,
            'rf_partial': pred.rf_partial,
            'categories_to_val_map': pred.categories_to_val_map
        }
        return pickle.dumps(state)

    @classmethod
    def deserialize(cls, bdb, binary):
        state = pickle.loads(binary)
        rf = cls(targets=state['targets'],
            conditions_numerical=state['conditions_numerical'],
            conditions_categorical=state['conditions_categorical'],
            rf_full=state['rf_full'], rf_partial=state['rf_partial'],
            categories_to_val_map=state['categories_to_val_map'])
        rf.prng = bdb.np_prng
        return rf

    @classmethod
    def name(cls):
        return 'random_forest'

    def __init__(self, targets=None, conditions_numerical=None,
            conditions_categorical=None, rf_full=None, rf_partial=None,
            categories_to_val_map=None):
        self.targets = targets
        self.conditions_numerical = conditions_numerical
        self.conditions_categorical = conditions_categorical
        if (conditions_numerical is not None
                and conditions_categorical is not None):
            self.conditions = self.conditions_numerical + \
                self.conditions_categorical
        self.rf_full = rf_full
        self.rf_partial = rf_partial
        self.categories_to_val_map = categories_to_val_map

    def train(self, df, targets, conditions):
        # Obtain the targets column.
        if len(targets) != 1:
            raise BLE(ValueError('RandomForest requires exactly one column in '
                'targets. Received {}'.format(targets)))
        if targets[0][1].lower() != 'categorical':
            raise BLE(ValueError('RandomForest can only classify CATEGORICAL '
                'columns. Received {}'.format(targets)))
        self.targets = [targets[0][0]]
        # Obtain the condition columns.
        if len(conditions) < 1:
            raise BLE(ValueError('RandomForest requires at least one column in '
                'conditions. Received {}'.format(conditions)))
        self.conditions_categorical = []
        self.conditions_numerical = []
        for c in conditions:
            if c[1].lower() == 'categorical':
                self.conditions_categorical.append(c[0])
            else:
                self.conditions_numerical.append(c[0])
        self.conditions = self.conditions_numerical + \
            self.conditions_categorical
        # The dataset.
        self.dataset = pd.DataFrame()
        # Lookup for categoricals to code.
        self.categories_to_val_map = dict()
        # Training set (regressors and labels)
        self.X_numerical = np.ndarray(0)
        self.X_categorical = np.ndarray(0)
        self.Y = np.ndarray(0)
        # Random Forests.
        self.rf_partial = RandomForestClassifier(n_estimators=100)
        self.rf_full = RandomForestClassifier(n_estimators=100)
        # Preprocess the data.
        self.dataset = utils.extract_sklearn_dataset(self.conditions,
            self.targets, df)
        self.categories_to_val_map = utils.build_categorical_to_value_map(
            self.conditions_categorical, self.dataset)
        self.X_categorical = utils.extract_sklearn_features_categorical(
            self.conditions_categorical, self.categories_to_val_map,
            self.dataset)
        self.X_numerical = utils.extract_sklearn_features_numerical(
            self.conditions_numerical, self.dataset)
        self.Y = utils.extract_sklearn_univariate_target(self.targets,
            self.dataset)
        # Train the random forest.
        self._train_rf()

    def _train_rf(self):
        """Trains the random forests classifiers.

        We train two classifiers, `partial` which is just trained on
        `conditions_numerical`, and `full` which is trained on
        `conditions_numerical+conditions_categorical`.

        This safe-guard feature is critical for querying; otherwise sklearn
        would crash whenever a categorical value unseen in training due to
        filtering (but existant in df nevertheless) was passed in.
        """
        # pylint: disable=no-member
        self.rf_partial.fit_transform(self.X_numerical, self.Y)
        self.rf_full.fit_transform(
            np.hstack((self.X_numerical, self.X_categorical)), self.Y)

    def _compute_targets_distribution(self, conditions):
        """Given conditions dict {feature_col:val}, returns the
        distribution and (class mapping for lookup) of the random label
        self.targets|conditions.
        """
        if not set(self.conditions).issubset(set(conditions.keys())):
            raise BLE(ValueError(
                'Must specify values for all the conditionals.\n'
                'Received: {}\n'
                'Expected: {}'.format(conditions, self.conditions_numerical +
                self.conditions_categorical)))

        # Are there any category values in conditions which never appeared during
        # training? If yes, we need to run the partial RF.
        unseen = any([conditions[cat] not in self.categories_to_val_map[cat]
            for cat in self.conditions_categorical])

        X_numerical = [conditions[col] for col in self.conditions_numerical]
        if unseen:
            distribution = self.rf_partial.predict_proba(X_numerical)
            classes = self.rf_partial.classes_
        else:
            X_categorical = [conditions[col] for col in
                self.conditions_categorical]
            X_categorical = utils.binarize_categorical_row(
                self.conditions_categorical, self.categories_to_val_map,
                X_categorical)
            distribution = self.rf_full.predict_proba(
                np.hstack((X_numerical, X_categorical)))
            classes = self.rf_partial.classes_
        return distribution[0], classes

    def simulate(self, n_samples, conditions):
        distribution, classes = self._compute_targets_distribution(conditions)
        draws = self.prng.multinomial(1, distribution, size=n_samples)
        return [classes[np.where(d==1)[0][0]] for d in draws]

    def logpdf(self, value, conditions):
        distribution, classes = self._compute_targets_distribution(conditions)
        if value not in classes:
            return -float('inf')
        return np.log(distribution[np.where(classes==value)[0][0]])



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Sphinx configuration file."""

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'numpydoc',
]

copyright = '2010-2016, MIT Probabilistic Computing Project'
master_doc = 'index'
project = 'bdbcontrib'
release = '0.1.dev'
version = '0.1'

nitpicky = True

numpydoc_show_class_members = False



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
def helpsub(pattern, replacement):
  """Set docstring = re.sub(pattern, replacement, original_docstring)."""
  assert pattern
  assert replacement is not None
  def _helpsub(fn):
    assert fn.__doc__
    fn.__doc__ = re.sub(pattern, replacement, fn.__doc__)
    return fn
  return _helpsub



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


###################################
# Decorators for library providers.

from collections import namedtuple
import inspect
import pydoc
import re
import types

from bayeslite.loggers import logged_query
from py_utils import helpsub
from population import Population

# This class is all about dynamically generated methods, so
# pylint: disable=no-member

PopulationTransformation = namedtuple(
    'PopulationTransformation',
    ['name', 'decorated_doc', 'method_doc', 'doc', 'transform'])

POPULATION_TRANSFORMATIONS = [
    PopulationTransformation(
        name='population',
        decorated_doc='''a bdbcontrib.population.Population''',
        method_doc=None,  # Is part of self, and should be hidden.
        doc='''Passes on the population object itself.''',
        transform=lambda pop: pop
        ),
    PopulationTransformation(
        name='interpret_bql',
        decorated_doc='''A BQL query (no %t or %g).''',
        method_doc='''A BQL query.
    You can also use %t and %g for the population-associated table name and
    the name of the population's generative-population-model-instance-generator,
    respectively.''',
        doc='''Fills in %t and %g in a BQL query.''',
        transform=lambda pop, arg: pop.interpret_bql(arg),
        ),
    PopulationTransformation(
        name='specifier_to_df',
        decorated_doc='''a pandas.DataFrame.''',
        method_doc='''a pandas.DataFrame or BQL query.''',
        doc='''Queries to create a dataframe from BQL if need be.''',
        transform=lambda pop, arg: pop.specifier_to_df(arg),
        ),
    PopulationTransformation(
        name='population_to_bdb',
        decorated_doc='''An active bayeslite.BayesDB instance.''',
        method_doc=None,
        doc='''Provides the population's BayesDB instance.''',
        transform=lambda pop: pop.bdb,
        ),
    PopulationTransformation(
        name='population_name',
        decorated_doc='''str
        the name of the relevant table in the bdb.''',
        method_doc=None,
        doc='''Provides the population's name.''',
        transform=lambda pop: pop.name,
        ),
    PopulationTransformation(
        name='generator_name',
        decorated_doc='''str
        the name of the relevant generator in the bdb.''',
        method_doc=None,
        doc='''Provides the population's generative population model's name.''',
        transform=lambda pop: pop.generator_name,
        ),
    PopulationTransformation(
        name='logger',
        decorated_doc='''A bayeslite.logger.BQLLogger instance.''',
        method_doc=None,
        doc='''Provides a logger to which one can .info, .warn, .plot, etc.''',
        transform=lambda pop: pop.logger,
        ),
    ]


ARGSPEC_TRANSFORMS = dict([(pt.name, pt.transform)
                           for pt in POPULATION_TRANSFORMATIONS])
METHOD_DOC_FILLERS = dict([('__'+pt.name+'__', pt.method_doc)
                           for pt in POPULATION_TRANSFORMATIONS])
DECORATED_DOC_FILLERS = dict([('__'+pt.name+'__', pt.decorated_doc)
                              for pt in POPULATION_TRANSFORMATIONS])
PT_DOC='\n'.join(["%s: %s\n\t%s" %
                  (pt.name, ("" if pt.method_doc is None else "[argspec]"),
                   pt.doc)
                  for pt in POPULATION_TRANSFORMATIONS])


def fill_documentation(docstr, fillers):
  if docstr is None:
    return None
  for key, explanation in fillers.iteritems():
    if explanation is None: # Remove it entirely.
      docstr = re.sub(r'\s*\w+\s*:\s*' + key, '\n', docstr, re.M)
    else:
      docstr = re.sub(key, explanation, docstr)
  return docstr

def redocument(fn, xfrms, fillers):
  title = 'wrapped as population.' + fn.__name__
  args, varargs, varkw, defaults = inspect.getargspec(fn)
  argspec = inspect.formatargspec(
      xfrms['required_names'] + xfrms['optional_names'],
      xfrms['varargs'], xfrms['varkw'], defaults,
      formatvalue=lambda x: '=' + repr(x))
  note = fill_documentation(fn.__doc__, fillers)
  first_nonempty_paragraph = []
  if note is None:
    note = ""
  else:
    first_nonempty_paragraph.append("")  # So join adds prior line break.
    lines = re.split(r'\n', note)
    for line in lines:
      line = line.strip()
      if line == "" and first_nonempty_paragraph[-1]:
        break
      if line != "":
        first_nonempty_paragraph.append(line)
    note = '\n\n' + note
  doc = title + argspec + note
  shortdoc = xfrms['name'] + argspec + "\n    ".join(first_nonempty_paragraph)
  return (doc, shortdoc)

def compile_argspec_transforms(fn, argspecs):
  '''Create a data structure to pass to apply_argspec_transforms.

  fn : the function that the argspecs should apply to.
  argspecs : argument specifiers as described in population_method.
  '''
  # There are two kinds of transforms: Ones that rely on just the population
  # object, and those that rely on both that and the data passed in for that
  # name.
  #
  # For those only in terms of the population: for optional arguments, supply
  # them if they are not present, passing through any value if they are.
  # Where an outside value is requested pass that in to the transformer as
  # specified.
  #
  # Positional arguments are similar, if they want just population, fill that
  # in and remove the associated argument. If they want a population and arg,
  # then satisfy that based on the default specified nearby.

  co_varnames, varargs, varkw, co_defaults = inspect.getargspec(fn)
  num_optional = 0 if co_defaults is None else len(co_defaults)
  num_required = len(co_varnames) - num_optional
  required_transformers = [None] * num_required
  optional_transformers = {}
  required_names = co_varnames[:num_required]
  optional_names = co_varnames[num_required:]
  for key, argspec in argspecs.iteritems():
    xform = ARGSPEC_TRANSFORMS[key]
    for argitem in (argspec if hasattr(argspec, '__iter__') else [argspec]):
      if isinstance(argitem, int):
        if argitem >= num_required or argitem < 0:
          raise IndexError(
            'Invalid positional specifier for %s: %d' % (fn.__name__, argitem))
        if required_transformers[argitem] is not None:
          raise IndexError(
            'Duplicate specifier for %s: %d' % (fn.__name__, argitem))
        required_transformers[argitem] = xform
      else:
        if argitem not in optional_names:
          raise IndexError(
            'No such optional argument for %s: %s' % (fn.__name__, argitem))
        if argitem in optional_transformers:
          raise IndexError(
            'Duplicate specifier for %s: %s' % (fn.__name__, argitem))
        population_only = (xform.__code__.co_argcount == 1)
        if population_only:
          def wrap_pop_only_opt(index, xfn): # To copy, not reference, free vars
            return lambda self, kwargs: \
              kwargs[index] if index in kwargs else xfn(self)
          newf = wrap_pop_only_opt(argitem, xform)
        else:
          def wrap_pop_arg_opt(xfn, index): # To copy, not reference, free vars
            return lambda self, kwargs: xfn(self, kwargs[index])
          newf = wrap_pop_arg_opt(xform, argitem)
        optional_transformers[argitem] = newf
  for name in optional_names:
    if name not in optional_transformers:
      def wrap_optarg_passthru(idx):
        return lambda _self, kwargs: kwargs[idx] if idx in kwargs else None
      optional_transformers[name] = wrap_optarg_passthru(name)


  wrapper_i = 0
  wrapped_transformers = []
  required_method_names = []
  for (decorated_i, xform) in enumerate(required_transformers):
    if xform is None: # No mention of this argument, so pass it through.
      def wrap_default(index): # To copy, not reference, free vars
        return lambda self, args: args[index]
      wrapped_transformers.append(wrap_default(wrapper_i))
      required_method_names.append(required_names[decorated_i])
      wrapper_i += 1
    else:
      population_only = (xform.__code__.co_argcount == 1)
      if population_only:
        def wrap_pop_only_req(xfn): # To copy, not reference, free vars.
          return lambda self, args: xfn(self)
        wrapped_transformers.append(wrap_pop_only_req(xform))
      else:
        def wrap_pop_arg_req(xfn, index): # To copy, not reference, free vars.
          return lambda self, args: xfn(self, args[index])
        wrapped_transformers.append(wrap_pop_arg_req(xform, wrapper_i))
        required_method_names.append(required_names[decorated_i])
        wrapper_i += 1
  return {'name': fn.__module__ + '.' + fn.__code__.co_name,
          'varargs': varargs,
          'varkw': varkw,
          'required_names': required_method_names,
          'required_transformers': wrapped_transformers,
          'optional_names': optional_names,
          'optional_transformers': optional_transformers,
          'co_defaults': co_defaults,
          }


def apply_argspec_transforms(pop, argspecs, args, kwargs):
  '''Apply the given transforms to the args and kwargs as specified.

     argspec: number, string, or list of numbers and strings.
         Not None. A positional argument number in args or a keyword argument
         name in kwargs.
     transforms: ([fn(pop, args)-->value], {'kwname': fn(pop, kwargs)-->value})
         The transformations to apply.
     args: tuple or list
         The args to be transformed. Note the lack of a star: just a list.
     kwargs: dict
         The kwargs to be transformed. Note the lack of stars: just a dict.

     Returns transformed (args, kwargs)
     '''
  required_names = argspecs['required_names']
  required_xforms = argspecs['required_transformers']
  optional_names = argspecs['optional_names']
  optional_xforms = argspecs['optional_transformers']
  co_defaults = argspecs['co_defaults']
  completed_args = list(args[:]) # mutable copy
  remaining_kwargs = kwargs.copy()

  # Check for duplicate args (to get the message right).
  all_names = required_names + optional_names
  for name in all_names[:len(args)]:
    if name in remaining_kwargs:
      raise TypeError("%s() got multiple values for keyword argument '%s'" %
                      (argspecs['name'], name))

  # Named required args:
  for name in required_names[len(args):]:
    if name in remaining_kwargs:
      completed_args.append(remaining_kwargs[name])
      del remaining_kwargs[name]
    else:
      raise TypeError("%s() takes at least %d arguments (%d given)" %
                      (argspecs['name'], len(required_names),
                       len(args) + len(kwargs)))

  new_varargs = []
  # Positional optional args:
  for opt_i in xrange(len(required_names), len(args)):
    value = args[opt_i]
    found = False
    for name in optional_names:
      if name not in remaining_kwargs:
        remaining_kwargs[name] = value
        found = True
        break
    if not found:
      if argspecs['varargs'] is not None:
        new_varargs.append(value)
      else:
        max_args = len(required_names) + len(optional_names)
        if max_args == 0:
          raise TypeError("%s() takes no arguments (%d given)" %
                          (argspecs['name'], len(args) + len(kwargs)))
        else:
          raise TypeError("%s() takes at most %d arguments (%d given)" %
                          (argspecs['name'], max_args, len(args) + len(kwargs)))

  new_args = []
  # Positional required args:
  for xform in required_xforms:
    new_args.append(xform(pop, completed_args))
  # Named optional args:
  for key in optional_names:
    xform = optional_xforms[key]
    new_args.append(xform(pop, remaining_kwargs))
    if key in remaining_kwargs:
      del remaining_kwargs[key]

  if argspecs['varkw'] is None:
    for key in remaining_kwargs:
      raise TypeError("%s() got an unexpected keyword argument '%s'" %
                      (argspecs['name'], key))

  # Allow the function's own defaults to override Nones:
  for i, optname in enumerate(optional_names):
      index = len(required_xforms) + i
      if new_args[index] is None:
          new_args[index] = co_defaults[i]

  return (new_args + new_varargs, remaining_kwargs)

@helpsub("__PT_DOC__", PT_DOC)
def population_method(**argspec_transforms):
  '''Create a method in the Population class to encapsulate this function.

Fill in documentation holes, but otherwise leave the function itself
untouched for library use.

For the types below, "arg" is a zero-indexed positional argument number
or a keyword argument name as a string. For keyword arguments, if the
caller provides a value, we leave it untouched. Positional arguments may
either be kept and transformed, or hidden from the argument list. If
the transform takes both a population and an arg, then it is kept and the
arg is transformed. If it takes only a population, then it is hidden (there
is no corresponding positional argument in the method, the documentation
line is removed, and other positional arguments shift to accommodate).

__PT_DOC__

In your method's docstring, use the above names with underscores to
get their appropriate documentation (if you want it replaced). For
example, use __population_name__ in place of the lines where you
would put the population name parameter's description, after its name.

For those transformations that do not have an [arg], if their documentation
line starts with the parameter name, then a colon, then the double
underscored transformation name, then the entire line is removed.
For example, a documentation line like
  table : __population_name__
would be entirely removed from the method docstring, because that argument
is filled in using the population instance, rather than being requested
from the caller.
     '''
  def decorator(fn):
    xfrms = compile_argspec_transforms(fn, argspec_transforms)
    def as_population_method(self, *args, **kwargs):
      with logged_query(query_string=fn.__code__.co_name,
                        bindings=(args, kwargs),
                        name=self.session_capture_name):
        self.check_representation()
        (dargs, dkwargs) = apply_argspec_transforms(self, xfrms, args, kwargs)
        result = None
        try:
          result = fn(*dargs, **dkwargs)
        except:
          self.logger.exception("")
          raise
        self.check_representation()
        return result

    (doc, shortdoc) = redocument(fn, xfrms, METHOD_DOC_FILLERS)
    as_population_method.__doc__ = doc
    Population.shortdocs.append(shortdoc)
    as_population_method.__name__ = xfrms['name']
    setattr(Population, fn.__code__.co_name, as_population_method)

    fn.__doc__ = fill_documentation(fn.__doc__, DECORATED_DOC_FILLERS)
    return fn

  return decorator



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import math
import pickle

import numpy as np
import pandas as pd

from sklearn.linear_model import LinearRegression

from bayeslite.exception import BayesLiteException as BLE
import bdbcontrib
from bdbcontrib.predictors import predictor
from bdbcontrib.predictors import sklearn_utils as utils

class MultipleRegression(predictor.IBayesDBForeignPredictor):
    """A linear regression foreign predictor.

    The `targets` must be a single numerical stattype.  The `conditions`
    may be arbitrary numerical or categorical columns.
    """

    @classmethod
    def create(cls, bdb, table, targets, conditions):
        cols = [c for c,_ in targets+conditions]
        df = bdbcontrib.bql_utils.table_to_df(bdb, table, cols)
        mr = cls()
        mr.train(df, targets, conditions)
        mr.prng = bdb.np_prng
        return mr

    @classmethod
    def serialize(cls, _bdb, pred):
        state = {
            'targets': pred.targets,
            'conditions_numerical': pred.conditions_numerical,
            'conditions_categorical': pred.conditions_categorical,
            'mr_full': pred.mr_full,
            'mr_partial': pred.mr_partial,
            'mr_full_noise': pred.mr_full_noise,
            'mr_partial_noise': pred.mr_partial_noise,
            'categories_to_val_map': pred.categories_to_val_map
        }
        return pickle.dumps(state)

    @classmethod
    def deserialize(cls, bdb, binary):
        state = pickle.loads(binary)
        mr = cls(targets=state['targets'],
            conditions_numerical=state['conditions_numerical'],
            conditions_categorical=state['conditions_categorical'],
            mr_full=state['mr_full'], mr_partial=state['mr_partial'],
            mr_full_noise=state['mr_full_noise'],
            mr_partial_noise=state['mr_partial_noise'],
            categories_to_val_map=state['categories_to_val_map'])
        mr.prng = bdb.np_prng
        return mr

    @classmethod
    def name(cls):
        return 'multiple_regression'

    def __init__(self, targets=None, conditions_numerical=None,
            conditions_categorical=None, mr_full=None, mr_partial=None,
            mr_full_noise=None, mr_partial_noise=None,
            categories_to_val_map=None):
        self.targets = targets
        self.conditions_numerical = conditions_numerical
        self.conditions_categorical = conditions_categorical
        if (self.conditions_numerical is not None
                and self.conditions_categorical is not None):
            self.conditions = self.conditions_numerical + \
                self.conditions_categorical
        self.mr_full = mr_full
        self.mr_partial = mr_partial
        self.mr_full_noise = mr_full_noise
        self.mr_partial_noise = mr_partial_noise
        self.categories_to_val_map = categories_to_val_map

    def train(self, df, targets, conditions):
        # Obtain the targets column.
        if len(targets) != 1:
            raise BLE(ValueError(
                'MultipleRegression requires at least one column '
                'in targets. Received {}'.format(targets)))
        if targets[0][1].lower() != 'numerical':
            raise BLE(ValueError(
                'MultipleRegression can only regress NUMERICAL '
                'columns. Received {}'.format(targets)))
        self.targets = [targets[0][0]]

        # Obtain the condition columns.
        if len(conditions) < 1:
            raise BLE(ValueError('MultipleRegression requires at least one '
                'column in conditions. Received {}'.format(conditions)))
        self.conditions_categorical = []
        self.conditions_numerical = []
        for c in conditions:
            if c[1].lower() == 'categorical':
                self.conditions_categorical.append(c[0])
            else:
                self.conditions_numerical.append(c[0])
        self.conditions = self.conditions_numerical + \
            self.conditions_categorical

        # The dataset.
        self.dataset = pd.DataFrame()
        # Lookup for categoricals to code.
        self.categories_to_val_map = dict()
        # Training set (regressors and labels)
        self.X_numerical = np.ndarray(0)
        self.X_categorical = np.ndarray(0)
        self.Y = np.ndarray(0)
        # Linear regressors.
        self.mr_partial = LinearRegression()
        self.mr_full = LinearRegression()

        # Preprocess the data.
        self.dataset = utils.extract_sklearn_dataset(self.conditions,
            self.targets, df)
        self.categories_to_val_map = utils.build_categorical_to_value_map(
            self.conditions_categorical, self.dataset)
        self.X_categorical = utils.extract_sklearn_features_categorical(
            self.conditions_categorical, self.categories_to_val_map,
            self.dataset)
        self.X_numerical = utils.extract_sklearn_features_numerical(
            self.conditions_numerical, self.dataset)
        self.Y = utils.extract_sklearn_univariate_target(self.targets,
            self.dataset)
        # Train the multiple regression.
        self._train_mr()

    def _train_mr(self):
        """Trains the regressions.

        We train two regressions, `partial` which is just trained on
        `conditions_numerical`, and `full` which is trained on
        `conditions_numerical+conditions_categorical`.

        This safe-guard feature is critical for querying; otherwise sklearn
        would crash whenever a categorical value unseen in training due to
        filtering (but existant in df nevertheless) was passed in.
        """
        self.mr_partial.fit(self.X_numerical, self.Y)
        self.mr_full.fit(np.hstack((self.X_numerical, self.X_categorical)),
            self.Y)

        self.mr_partial_noise = \
            np.linalg.norm(self.Y-self.mr_partial.predict(
                self.X_numerical))/len(self.Y)

        self.mr_full_noise = \
            np.linalg.norm(self.Y-self.mr_full.predict(
                np.hstack((self.X_numerical, self.X_categorical))))/len(self.Y)

    def _compute_targets_distribution(self, conditions):
        """Given conditions dict {feature_col:val}, returns the conditional
        mean of the `targets`, and the scale of the Gaussian noise.
        """
        if not set(self.conditions).issubset(set(conditions.keys())):
            raise BLE(ValueError(
                'Must specify values for all the conditionals.\n'
                'Received: {}\n'
                'Expected: {}'.format(conditions, self.conditions_numerical +
                self.conditions_categorical)))

        # Are there any category values in conditions which never appeared during
        # training? If yes, we need to run the partial RF.
        unseen = any([conditions[cat] not in self.categories_to_val_map[cat]
            for cat in self.conditions_categorical])

        X_numerical = [conditions[col] for col in self.conditions_numerical]

        if unseen:
            inputs = np.array([X_numerical])
            assert inputs.shape == (1, len(self.conditions_numerical))
            predictions = self.mr_partial.predict(inputs)
            noise = self.mr_partial_noise
        else:
            X_categorical = [conditions[col] for col in
                self.conditions_categorical]
            X_categorical = utils.binarize_categorical_row(
                self.conditions_categorical, self.categories_to_val_map,
                X_categorical)
            inputs = np.concatenate(([X_numerical], [X_categorical]), axis=1)
            assert inputs.shape == \
                (1, len(self.conditions_numerical) + len(X_categorical))
            predictions = self.mr_full.predict(inputs)
            noise = self.mr_full_noise

        return predictions[0], noise

    def simulate(self, n_samples, conditions):
        prediction, noise = self._compute_targets_distribution(conditions)
        return list(prediction + self.prng.normal(scale=noise, size=n_samples))

    def logpdf(self, value, conditions):
        prediction, noise = self._compute_targets_distribution(conditions)
        return logpdfGaussian(value, prediction, noise)

HALF_LOG2PI = 0.5 * math.log(2 * math.pi)
def logpdfGaussian(x, mu, sigma):
    deviation = x - mu
    return - math.log(sigma) - HALF_LOG2PI \
        - (0.5 * deviation * deviation / (sigma * sigma))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import matplotlib
matplotlib.use('Agg')

import argparse
import cPickle as pickle # json doesn't like tuple dict keys
import logging
import re

from bdbcontrib.experiments.probe import log
from bdbcontrib.experiments.visualization import plot_results

def analysis_count_from_file_name(fname):
    return int(re.match(r'.*-[0-9]*m-([0-9]*)i.bdb', fname).group(1))

def plot_probe_results(filename, outdir):
    log("Loading probe results from %s" % filename)
    with open(filename, "r") as f:
        results = pickle.load(f)
    results = [((probe, n_models, analysis_count_from_file_name(fname)), val)
               for ((fname, n_models, probe), val) in results.iteritems()]
    plot_results(results, outdir=outdir)

parser = argparse.ArgumentParser(
    description='Process a collection of saved probe results')
parser.add_argument('-i', '--infile',
                    help="Read probe results from the given file.")
parser.add_argument('-o', '--outdir',
                    help="Write polts to the given output directory.")

def main():
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)
    plot_probe_results(args.infile, args.outdir)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

try:
    from setuptools import setup
    from setuptools.command.build_py import build_py
    from setuptools.command.sdist import sdist
    from setuptools.command.test import test as TestCommand
except ImportError:
    from distutils.core import setup
    from distutils.cmd import Command
    from distutils.command.build_py import build_py
    from distutils.command.sdist import sdist

    class TestCommand(Command):
        user_options = []
        def initialize_options(self): pass
        def finalize_options(self): pass
        def run(self): self.run_tests()

def get_version():
    with open('VERSION', 'rb') as f:
        version = f.read().strip()

    # Append the Git commit id if this is a development version.
    if version.endswith('+'):
        import re
        import subprocess
        version = version[:-1]
        tag = 'v' + version
        desc = subprocess.check_output([
            'git', 'describe', '--dirty', '--long', '--match', tag,
        ])
        match = re.match(r'^v([^-]*)-([0-9]+)-(.*)$', desc)
        assert match is not None
        verpart, revpart, localpart = match.groups()
        assert verpart == version
        # Local part may be g0123abcd or g0123abcd-dirty.  Hyphens are
        # not kosher here, so replace by dots.
        localpart = localpart.replace('-', '.')
        full_version = '%s.post%s+%s' % (verpart, revpart, localpart)
    else:
        full_version = version

    # Strip the local part if there is one, to appease pkg_resources,
    # which handles only PEP 386, not PEP 440.
    if '+' in full_version:
        pkg_version = full_version[:full_version.find('+')]
    else:
        pkg_version = full_version

    # Sanity-check the result.  XXX Consider checking the full PEP 386
    # and PEP 440 regular expressions here?
    assert '-' not in full_version, '%r' % (full_version,)
    assert '-' not in pkg_version, '%r' % (pkg_version,)
    assert '+' not in pkg_version, '%r' % (pkg_version,)

    return pkg_version, full_version

pkg_version, full_version = get_version()

def write_version_py(path):
    try:
        with open(path, 'rb') as f:
            version_old = f.read()
    except IOError:
        version_old = None
    version_new = '__version__ = %r\n' % (full_version,)
    if version_old != version_new:
        print 'writing %s' % (path,)
        with open(path, 'wb') as f:
            f.write(version_new)

from distutils.command.install import INSTALL_SCHEMES
for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']
example_files = {}
import os
import re
with open('MANIFEST.in', 'r') as manifest:
    for line in manifest:
        line = line.strip()
        if line.startswith('include examples'):
            line = line[len("include "):]
            dirname = os.path.join('bdbcontrib', os.path.dirname(line))
            if dirname not in example_files:
                example_files[dirname] = []
            example_files[dirname].append(line)

class local_build_py(build_py):
    def run(self):
        write_version_py(version_py)
        build_py.run(self)

# Make sure the VERSION file in the sdist is exactly specified, even
# if it is a development version, so that we do not need to run git to
# discover it -- which won't work because there's no .git directory in
# the sdist.
class local_sdist(sdist):
    def make_release_tree(self, base_dir, files):
        import os
        sdist.make_release_tree(self, base_dir, files)
        version_file = os.path.join(base_dir, 'VERSION')
        print('updating %s' % (version_file,))
        # Write to temporary file first and rename over permanent not
        # just to avoid atomicity issues (not likely an issue since if
        # interrupted the whole sdist directory is only partially
        # written) but because the upstream sdist may have made a hard
        # link, so overwriting in place will edit the source tree.
        with open(version_file + '.tmp', 'wb') as f:
            f.write('%s\n' % (pkg_version,))
        os.rename(version_file + '.tmp', version_file)

# XXX Several horrible kludges here to make `python setup.py test' work:
#
# - Standard setputools test command searches for unittest, not
#   pytest.
#
# - pytest suggested copypasta assumes . works in sys.path; we
#   deliberately make . not work in sys.path and require ./build/lib
#   instead, in order to force a clean build.
#
# - Must set PYTHONPATH too for shell tests, which fork and exec a
#   subprocess which inherits PYTHONPATH but not sys.path.
#
# - build command's build_lib variable is relative to source
#   directory, so we must assume os.getcwd() gives that.
#
# (The shell test issues are irrelevant for the moment in bdbcontrib,
# but there's no harm in setting PYTHONPATH anyway.)
class cmd_pytest(TestCommand):
    def __init__(self, *args, **kwargs):
        TestCommand.__init__(self, *args, **kwargs)
        self.test_suite = 'tests'
        self.build_lib = None
    def finalize_options(self):
        TestCommand.finalize_options(self)
        # self.build_lib = ...
        self.set_undefined_options('build', ('build_lib', 'build_lib'))
    def run_tests(self):
        import pytest
        import os
        import os.path
        import sys
        sys.path = [os.path.join(os.getcwd(), self.build_lib)] + sys.path
        os.environ['BAYESDB_WIZARD_MODE'] = '1'
        os.environ['BAYESDB_DISABLE_VERSION_CHECK'] = '1'
        os.environ['PYTHONPATH'] = ':'.join(sys.path)
        sys.exit(pytest.main(['tests']))

# XXX These should be attributes of `setup', but helpful distutils
# doesn't pass them through when it doesn't know about them a priori.
version_py = 'src/version.py'

setup(
    name='bdbcontrib',
    version=pkg_version,
    description='Hodgepodge library of extras for bayeslite',
    url='http://probcomp.csail.mit.edu/bayesdb',
    author='MIT Probabilistic Computing Project',
    author_email='bayesdb@mit.edu',
    license='Apache License, Version 2.0',
    install_requires=[
        'bayeslite>=0.1.8',
        'ipython[notebook]>=3',
        'markdown2',
        'matplotlib',
        'numpy',
        'numpydoc',
        'pandas>=0.17.0',
        'seaborn>=0.7',  # seaborn 0.6 is at the root of #94.
        'scipy',
        'sklearn',
        'sklearn-pandas',
        'sphinx',
        'tornado>=4.0',
    ],
    tests_require=[
        'flaky',  # for bdb-experiments
        'mock',
        'pillow',
        'pytest',
    ],
    packages=[
        'bdbcontrib',
        'bdbcontrib.predictors',
        'bdbcontrib.metamodels',
        'bdbcontrib.experiments',
    ],
    package_dir={
        'bdbcontrib': 'src',
    },
    data_files=example_files.items(),
    scripts=[
        'scripts/bayesdb-demo',
    ],
    cmdclass={
        'build_py': local_build_py,
        'sdist': local_sdist,
        'test': cmd_pytest,
    },
)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt

import bayeslite
import os
import pandas as pd
import random
import cStringIO as StringIO

from bayeslite.read_pandas import bayesdb_read_pandas_df
from bdbcontrib.crosscat_utils import draw_state
from crosscat.utils import data_utils as du

def draw_a_cc_state(filename):
    rng_seed = random.randrange(10000)
    num_rows = 100
    num_cols = 50
    num_splits = 5
    num_clusters = 5

    nan_prop = .25

    table_name = 'plottest'
    generator_name = 'plottest_cc'

    # generate some clustered data
    ccmd = du.generate_clean_state(rng_seed, num_clusters, num_cols, num_rows,
                                   num_splits)
    T, _M_c, _M_r, _X_L, _X_D = ccmd

    for row in range(num_rows):
        for col in range(num_cols):
            if random.random() < nan_prop:
                T[row][col] = float('nan')

    input_df = pd.DataFrame(T, columns=['col_%i' % i for i in range(num_cols)])

    os.environ['BAYESDB_WIZARD_MODE']='1'
    bdb = bayeslite.bayesdb_open()
    bayesdb_read_pandas_df(bdb, table_name, input_df, create=True)
    bdb.execute('''
        create generator {} for {} using crosscat(guess(*))
    '''.format(generator_name, table_name))
    bdb.execute('initialize 4 models for {}'.format(generator_name))
    bdb.execute('analyze {} for 10 iterations wait'.format(generator_name))
    plt.figure(facecolor='white', tight_layout=False)
    draw_state(bdb, 'plottest', 'plottest_cc', 0,
               separator_width=1, separator_color=(0., 0., 1., 1.),
               short_names=False, nan_color=(1, .15, .25, 1.))
    plt.savefig(filename)

def test_draw_cc_smoke():
    f = StringIO.StringIO()
    draw_a_cc_state(f)
    assert len(f.getvalue()) > 1000

# For manually inspecting the generated figure.
if __name__ == '__main__':
    draw_a_cc_state('state.png')
    print "Figure saved to 'state.png'"



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import markdown2
import os
import shutil
import shlex

from bayeslite.shell.hook import bayesdb_shell_cmd
from bayeslite.shell.pretty import pp_list

import bdbcontrib
import bdbcontrib.shell_utils as utils


ROOTDIR = os.path.dirname(os.path.abspath(__file__))
READTOHTML_CSS = os.path.join(ROOTDIR, 'readtohtml.css')


@bayesdb_shell_cmd('readtohtml')
def render_bql_as_html(self, argin):
    """read BQL file and output to HTML and markdown
    <bql_file> <output_directory>

    Example:
    bayeslite> .readtohtml myscript.bql analyses/myanalsis
    """
    parser = utils.ArgumentParser(prog='.readtohtml')
    parser.add_argument('bql_file', metavar='bql-file', type=str,
        help='Name of the file containing the bql script.')
    parser.add_argument('output_dir', metavar='output-dir', type=str,
        help='Name of the output directory.')

    try:
        args = parser.parse_args(shlex.split(argin))
    except utils.ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    bql_file = args.bql_file
    output_dir = args.output_dir


    head = '''
    <html>
        <head>
            <link rel="stylesheet" type="text/css" href="style.css">
        </head>
    '''

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    with open(bql_file) as f:
        mdstr = utils.mdread(f, output_dir, self)

    html = head + markdown2.markdown(mdstr) + '</html>'

    htmlfilename = os.path.splitext(os.path.basename(bql_file))[0]
    htmlfilename = os.path.join(output_dir, htmlfilename + '.html')
    with open(htmlfilename, 'w') as f:
        f.write(html)
    shutil.copyfile(READTOHTML_CSS, os.path.join(output_dir, 'style.css'))

    # Because you want to know when it's done:
    self.stdout.write(utils.unicorn())
    self.stdout.write("Output saved to %s\n" % (output_dir,))


@bayesdb_shell_cmd('nullify')
def nullify(self, argin):
    """replace user-specified missing value with NULL
    <table> <value>

    Example:
    bayeslite> .nullify mytable NaN
    bayeslite> .nullify mytable ''
    """
    parser = utils.ArgumentParser(prog='.nullify')
    parser.add_argument('table', type=str,
        help='Name of the table.')
    parser.add_argument('value', type=str,
        help='Target string to nullify.')

    try:
        args = parser.parse_args(shlex.split(argin))
    except utils.ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    bdbcontrib.nullify(self._bdb, args.table, args.value)


@bayesdb_shell_cmd('cardinality')
def cardinality(self, argin):
    """show cardinality of columns in table
    <table> [<column> <column> ...]

    Example:
    bayeslite> .cardinality mytable
    bayeslite> .cardinality mytable col1 col2 col3
    """
    parser = utils.ArgumentParser(prog='.cardinality')
    parser.add_argument('table', type=str,
        help='Name of the table.')
    parser.add_argument('cols', type=str, nargs='*',
        help='Target columns for which to compute cardinality.')

    try:
        args = parser.parse_args(shlex.split(argin))
    except utils.ArgparseError as e:
        self.stdout.write('%s' % (e.message,))
        return

    counts = bdbcontrib.cardinality(self._bdb, args.table, cols=args.cols)
    pp_list(self.stdout, counts, ['column', 'cardinality'])



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Visualize results of probing across iteration and model count."""

import os
import string
import warnings

import pandas as pd
import matplotlib.pylab as plt
import seaborn as sns

from bdbcontrib.experiments.probe import log

# results :: [((probe_name, model_ct, iters), tagged aggregated value)]

def plot_results(results, outdir="figures", ext=".png"):
    """Plot the aggregate results of probing.

    `results` is a list of pairs giving probe conditions and
    aggregated probe results.

    `outdir` is the name of a directory to which to write the visualizations.
    Default: "figures".

    `ext` is the file extension for visualizations, which determines
    the image format used.  Default ".png".

    Each probe condition is expected to be a 3-tuple: probe name,
    model count, analysis iteration count.  Each result is expected to
    be a tagged aggregate (see aggregation.py).

    Each numerical probe produces one plot, named after the probe.
    The plot facets over the model count, displays the iteration count
    on the x-axis, and a violin plot of the results on the y axis.

    All boolean probes are aggregated into one plot named
    "boolean-probes", whose y axis is the frequency of a "True"
    result.  Each probe is a line giving the relationship of the
    frequency to the number of analysis iterations.

    """
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    replications = num_replications(results)
    probes = sorted(set((pname, ptype)
                        for ((pname, _, _), (ptype, _)) in results))
    for probe, ptype in probes:
        if not ptype == 'num': continue
        grid = plot_results_numerical(results, probe)
        grid.fig.suptitle(probe + ", %d replications" % replications)
        # XXX Actually shell quote the probe name
        figname = string.replace(probe, " ", "-").replace("/", "") + ext
        savepath = os.path.join(outdir, figname)
        grid.savefig(savepath)
        plt.close(grid.fig)
        log("Probe '%s' results saved to %s" % (probe, savepath))
    grid = plot_results_boolean(results)
    grid.fig.suptitle("Boolean probes, %d replications" % replications)
    figname = "boolean-probes" + ext
    savepath = os.path.join(outdir, figname)
    grid.savefig(savepath)
    plt.close(grid.fig)
    log("Boolean probe results saved to %s" % (savepath,))

def plot_results_numerical(results, probe):
    # results :: [((probe_name, model_ct, iters), tagged result_set)]
    data = ((iters, model_ct, value)
            for ((pname, model_ct, iters), (ptype, values)) in results
        if pname == probe and ptype == 'num'
        for value in values)
    cols = ["num iterations", "n_models", "value"]
    df = pd.DataFrame.from_records(data, columns=cols) \
                     .sort(["num iterations", "n_models"])
    g = sns.FacetGrid(df, col="n_models", size=5, col_wrap=3)
    g.map(sns.violinplot, "num iterations", "value")
    return g

def plot_results_boolean(results):
    # results :: [((probe_name, model_ct, iters), tagged result_set)]
    data = ((iters, model_ct, pname, float(t)/(t+f))
            for ((pname, model_ct, iters), (ptype, value)) in results
        if ptype == 'bool'
        for (t, f) in [value])
    cols = ["num iterations", "n_models", "probe", "freq"]
    df = pd.DataFrame.from_records(data, columns=cols) \
                     .sort(["probe", "num iterations", "n_models"])
    g = sns.FacetGrid(df, col="n_models", hue="probe", size=5, col_wrap=3)
    g.map(plt.plot, "num iterations", "freq").add_legend()
    return g

def num_replications(results):
    replication_counts = \
        set((len(l) for (_, (ptype, l)) in results
             if ptype == 'num')).union(
        set((t+f for (_, (ptype, l)) in results
             if ptype == 'bool'
             for (t, f) in [l])))
    replication_count = next(iter(replication_counts))
    if len(replication_counts) > 1:
        msg = "Non-rectangular results found; replication counts range " \
              "over %s, using %s" % (replication_counts, replication_count)
        warnings.warn(msg)
    return replication_count



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from contextlib import contextmanager
import json
import os
import re
import subprocess
import sys
import traceback

def get_out_and_err(notebook_path=None, outfile=None, errfile=None):
  output = None
  error = None
  if notebook_path is not None:
    cmd = 'runipy --matplotlib --stdout "%s.ipynb"' % (notebook_path,)
    p = subprocess.Popen(cmd, shell=True, stdin=None,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         close_fds=True)
    output = p.stdout.read()
    error = p.stderr.read()
  elif outfile is not None and errfile is not None:
    with open(outfile, 'r') as out:
      output = out.read()
    with open(errfile, 'r') as err:
      error = err.read()
  elif len(sys.argv) == 3:
    return get_out_and_err(sys.argv[1], sys.argv[2])
  else:
    raise ValueError("Specify a notebook path or out and err files to verify.")
  return (output, error)

def ident(cell_i, cell, msglimit=None):
  return "Cell[%s]: %s" % (cell_i, repr(cell)[:msglimit])

def assert_markdown_matches(rxp, **kwargs):
  def assert_markdown_matches_tester(cell_i, cell, dscr):
    assert cell['cell_type'] in ('markdown', 'heading'), \
        "Expected markdown cell: " + dscr
    assert 'source' in cell, dscr
    full_source = '\n'.join(cell['source'])
    assert re.search(rxp, full_source, **kwargs), \
        "Expected /%s/\nFound: %s\nFor cell: %s" % (rxp, full_source, dscr)
    return True
  return assert_markdown_matches_tester

def assert_markdown_not(rxp, **kwargs):
  def assert_markdown_not_matches_tester(cell_i, cell, dscr):
    assert cell['cell_type'] in ('markdown', 'heading'), \
        "Expected markdown cell: " + dscr
    assert 'source' in cell, dscr
    full_source = '\n'.join(cell['source'])
    assert not re.search(rxp, full_source, **kwargs), \
        ("Expected not to find /%s/\nFound: %s\nFor cell: %s" %
         (rxp, full_source, dscr))
    return True
  return assert_markdown_not_matches_tester

def make_warning_tester(rxp, optional=False, **kwargs):
  def assert_warns_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert 'outputs' in cell, dscr
    found = False
    for output in cell['outputs']:
      assert 'output_type' in output, dscr
      if 'stream' != output['output_type']:
        continue
      assert 'text' in output, dscr
      for msg_i, msg in enumerate(output['text']):
        if re.search(r'/(.*):\d+:.*(warn(ings?)?)\W', msg, re.I):
          if re.search(rxp, msg, **kwargs):
            # Found it!
            # Mark it so we don't die when checking for unchecked warnings:
            output['text'][msg_i] += ' __verify_notebook_warning_ok__'
            return True
    assert optional or found, \
      "No matching warning: /%s/ in cell %s" % (rxp, dscr)
    return True
  return assert_warns_tester

def assert_warns(rxp, **kwargs):
  return make_warning_tester(rxp, optional=False, **kwargs)

def allow_warns(rxp, **kwargs):
  return make_warning_tester(rxp, optional=True, **kwargs)

def assert_raises(rxp, **kwargs):
  def assert_raises_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert 'outputs' in cell, dscr
    found = False
    for output in cell['outputs']:
      assert 'output_type' in output, dscr
      if 'pyerr' != output['output_type']:
        continue
      assert 'evalue' in output, dscr
      if re.search(rxp, output['evalue'], **kwargs):
        # Found it!
        # Mark it so we don't die when checking for unchecked warnings:
        output['evalue'] += ' __verify_notebook_error_ok__'
        return True
    assert found, "No matching error: /%s/ in cell %s" % (rxp, dscr)
  return assert_raises_tester

def assert_no_pyout():
  def assert_no_pyout_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert ('outputs' not in cell or
            [] == cell['outputs']), "Expected no output from cell: " + dscr
    return True
  return assert_no_pyout_tester

def assert_stream_matches(rxp, **kwargs):
  def assert_stream_matches_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert 'outputs' in cell, "No outputs in cell: " + dscr
    found = False
    for output in cell['outputs']:
      assert 'output_type' in output, "No output type: " + dscr
      if 'stream' != output['output_type']:
        continue
      assert 'text' in output, "No text in output: " + dscr
      for msg_i, msg in enumerate(output['text']):
        if re.search(rxp, msg, **kwargs):
          return True
    assert found, "No matching stream text: /%s/ in cell %s" % (rxp, dscr)
  return assert_stream_matches_tester

def assert_stream_not(rxp, **kwargs):
  def assert_stream_not_matches_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert 'outputs' in cell, "No outputs in code cell: " + dscr
    for output in cell['outputs']:
      assert 'output_type' in output, "No output_type in cell: " + dscr
      if 'stream' != output['output_type']:
        continue
      assert 'text' in output, "No text in stream: " + dscr
      for msg_i, msg in enumerate(output['text']):
        assert not re.search(rxp, msg, **kwargs), \
          "Expected not to find /%s/ but found it in cell: %s" % (rxp, dscr)
    return True
  return assert_stream_not_matches_tester

def assert_pyout_matches(rxp, **kwargs):
  def assert_pyout_matches_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert 'outputs' in cell, "No outputs in cell: " + dscr
    found = False
    for output in cell['outputs']:
      assert 'output_type' in output, "No output type: " + dscr
      if 'pyout' != output['output_type']:
        continue
      assert 'text' in output, "No text in output: " + dscr
      for msg_i, msg in enumerate(output['text']):
        if re.search(rxp, msg, **kwargs):
          return True
    assert found, "No matching output: /%s/ in cell %s" % (rxp, dscr)
  return assert_pyout_matches_tester

def assert_pyout_not(rxp, **kwargs):
  def assert_pyout_not_matches_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert 'outputs' in cell, "No outputs in code cell: " + dscr
    for output in cell['outputs']:
      assert 'output_type' in output, "No output_type in cell: " + dscr
      if 'pyout' != output['output_type']:
        continue
      assert 'text' in output, "No text in pyout: " + dscr
      for msg_i, msg in enumerate(output['text']):
        assert not re.search(rxp, msg, **kwargs), \
          "Expected not to find /%s/ but found it in cell: %s" % (rxp, dscr)
    return True
  return assert_pyout_not_matches_tester

def assert_has_png():
  def assert_has_png_tester(cell_i, cell, dscr):
    assert 'code' == cell['cell_type'], "Expected code cell: " + dscr
    assert 'outputs' in cell, dscr
    found = False
    for output in cell['outputs']:
      assert 'output_type' in output, dscr
      if 'display_data' == output['output_type'] and 'png' in output:
        output['png'] = '__verify_notebook_png_ok__'
        return True
    assert found, "Expected png for cell: " + dscr
  return assert_has_png_tester


def check_results(results,
                  exhaustive=None,
                  required=None,
                  callback=None,
                  warnings_are_errors=True,
                  msglimit=None,
                  errstream=sys.stderr):
  """Check that the ipynb ran successfully and that they match given patterns.

  exhaustive: [ (input cell regex, [ ANDed output cell testers ] ) ]
      * In the same order as the notebook,
      * All cells must be listed,
      * A cell tester is a fn(cell_i, cell, description) --> boolean
      * If a cell tester is not callable, it is assumed to be a regex r that
        is meant to be short for assert_markdown_matches or assert_pyout_matches
        as appropriate for the cell type.
      * It may raise or return False or None to fail, must return True to pass.
      * An empty cell pattern list is okay (checks success as usual).
      * Note the helpers defined in this module:
        + assert_markdown_matches(rxp, **kwargs_to_re_search)
        + assert_markdown_not(rxp, **kwargs_to_re_search)
        + assert_warns(rxp, **kwargs_to_re_search)
        + assert_raises(rxp, **kwargs_to_re_search)
        + assert_no_pyout()
        + assert_pyout_matches(rxp, **kwargs_to_re_search)
        + assert_pyout_not(rxp, **kwargs_to_re_search)
        + assert_has_png()
      * If you use assert_warns or assert_raises, then the messages found by
        those will not cause the overall correctness check to fail.

  required: [ (input cell regex, [ ANDed output cell testers ] ) ]
      * As before, all given testers must pass for a cell that matches the
        input cell regex, but
      * Not all notebook cells need be listed, and they need not be in order.
      * Even so, each input regex can still only match once, and
      * All input regexes must still match.

  callback: fn(cell_i, cell, description) --> boolean
      * Called on each cell.
      * Like testers above, may raise, or return False or None to fail,
        and must return True to pass.

  warnings_are_errors: boolean (default=True)
      Fail if a cell issues a warning (except as marked by assert_warns).

  msglimit: int
      Show this many characters of a notebook cell in an error message. Useful
      to make errors on very long cells more readable.
      Example: msglimit=(None if pytest.config.options.verbose else 1000)

  errstream: StringIO (default sys.stderr) or list or None.
      Where to write individual AssertionErrors (other errors propagate).
      Pass a list to collect them and let you do what you please.
      Pass None to raise immediately rather than collecting all failures.
  """
  exh = exhaustive[:] if exhaustive is not None else []  # Mutable copy.
  req = required[:] if required is not None else []  # Mutable copy.
  (output, error) = results
  try:
    notebook = json.loads(output)
  except:
    print error
    raise
  cells = notebook['worksheets'][0]['cells']

  errors = []
  @contextmanager
  def each_error():
    try:
      yield
    except AssertionError as e:
      if errstream is None:
        raise
      elif isinstance(errstream, list):
        # Pylint can't tell that isinstance above protected errstream from
        # being a file.
        errstream.append(e)  # pylint: disable=no-member
      else:
        errors.append(e)
        traceback.print_exc(limit=1, file=errstream)
        print >>errstream, "\n"

  for cell_i, cell in enumerate(cells):
    with each_error():
      dscr = "Cell[%s]: %s" % (cell_i, repr(cell)[:msglimit])
      assert cell['cell_type'] in ('markdown', 'heading', 'code'), dscr
      if callback:
        assert callback(cell_i, cell, dscr)

      # Find the tests to run:
      input = '\n'.join(cell['input'] if cell['cell_type'] == 'code'
                        else cell['source'])
      cell_tests = []
      if exhaustive is not None:
        assert cell_i < len(exh), "Unexpected cell at end: " + dscr
        spec = exh[cell_i]
        assert 2 == len(spec), \
            ("Expected an input pattern and a list of tests,\n"
             "Found %s\nFor cell %s" % (spec, dscr))
        pttn, cell_tests = spec
        assert re.search(pttn, input), \
          "Expected /%s/\nFound: %s\nFor cell: %s" % (pttn, input, dscr)

      for i, spec in enumerate(req):
        assert 2 == len(spec), \
            ("Expected an input pattern and a list of tests,\n"
             "Found %s\nFor cell %s" % (spec, dscr))
        pttn, tests = spec
        if re.search(pttn, input):
          cell_tests += tests
          req.pop(i)

      for test in cell_tests:
        if not hasattr(test, '__call__'):
          if cell['cell_type'] == 'code':
            test = assert_pyout_matches(test)
          else:
            test = assert_markdown_matches(test)
        ans = test(cell_i, cell, dscr)
        assert ans, "Tester %r returned %r on cell: %s" % (test, ans, dscr)

      # Standard tests:
      if cell['cell_type'] == 'code':
        for output in cell['outputs']:
          assert output['output_type'] in (
            'pyerr', 'stream', 'pyout', 'display_data'), dscr
          if (output['output_type'] == 'pyerr'):
            assert 'evalue' in output
            full_output = output['evalue']
            assert re.search('__verify_notebook_error_ok__', full_output), dscr
          elif output['output_type'] == 'stream':
            full_output = '\n'.join(output['text'])
            if (warnings_are_errors and
                re.search(r'/(.*):\d+:.*(warn(ings?)?)\W', full_output, re.I)):
              assert re.search(r'__verify_notebook_warning_ok__', full_output),\
                  "Undeclared warning [%s] in cell: %s" % (full_output, dscr)
          elif output['output_type'] == 'display_data':
            if exhaustive is not None and 'png' in output:
              assert '__verify_notebook_png_ok__' == output['png'], \
                "Undeclared png in cell: " + dscr
          else:
            pass  # No automatic tests for successful code.
      else:
        pass  # No automatic tests for markdown

  if callback:
    with each_error():
      assert callback(None, None, "At end.")

  for spec in exh[len(cells):] + req:
    with each_error():
      assert 2 == len(spec), \
          ("Expected an input pattern and a list of tests,\n"
           "Found %s\nFor cell %s" % (spec, dscr))
      pttn, _test = spec
      assert False, "Expected pattern /%s/ not found" % (pttn,)

  assert 0 == len(errors), "See the error stream for individual error reports."

def run_and_verify_notebook(notebook_path_basename, **kwargs):
  '''runipy it and verify it. Adds ".ipynb" to the path.

  If the DEBUG_TESTS environment variable is set, cache notebook results in
  basename.out-ipynb and basename.err. That avoids rerunning the notebook
  when first writing tests for an existing notebook, or when debugging the
  tests themselves.
  '''
  if "DEBUG_TESTS" not in os.environ:
    check_results(get_out_and_err(notebook_path_basename), **kwargs)
  else:
    outfile=notebook_path_basename + ".out-ipynb"
    errfile=notebook_path_basename + ".err"
    if os.path.exists(outfile) and os.path.exists(errfile):
      check_results(get_out_and_err(outfile=outfile, errfile=errfile), **kwargs)
    else:
      (out, err) = get_out_and_err(notebook_path_basename)
      with open(outfile, 'w') as outfd:
        outfd.write(out)
      with open(errfile, 'w') as errfd:
        errfd.write(err)
      check_results((out, err), **kwargs)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import argparse
import os
import string
from cStringIO import StringIO


PLOTTING_COMMANDS = ['.heatmap', '.histogram', '.show',
                     '.chainplot', '.ccstate', '.bar',
                     '.mihist']
BPROMPT = 'bayeslite> '
CPROMPT = '      ...> '


# Kludgey workaround for idiotic Python argparse module which exits
# the process on argument parsing failure.  We override the exit
# method of ArgumentParser so that it raises an exception instead of
# exiting the process, which we then catch around parser.parse_args()
# in order to report the message and return to the command loop.
class ArgparseError(Exception):
    def __init__(self, status, message):
        self.status = status
        self.message = message


class ArgumentParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        raise ArgparseError(status, message)

def is_dot_command(line):
    if is_blank_line(line):
        return False
    return line[0] == '.'


def is_continuation(line):
    if is_blank_line(line):
        return False
    return line[0] in string.whitespace


def is_plotting_command(line):
    if not is_dot_command(line):
        return False
    return line.split()[0] in PLOTTING_COMMANDS


def is_blank_line(line):
    return len(line) == 0 or line.isspace()


def is_comment(line):
    if len(line) < 2:
        return False
    return line[:2] == '--'


def get_line_type(line):
    if is_comment(line):
        return 'comment'
    elif is_blank_line(line):
        return 'blank'
    else:
        return 'code'


def do_and_cap(shell, cmd):
    assert len(cmd) > 0
    shell.stderr.write("Performing ")
    shell.stderr.write(cmd)
    shell.stderr.write("\n")
    shell.stderr.flush()

    backup = shell.stdout
    stream = StringIO()
    shell.stdout = stream

    shell.onecmd(cmd)

    shell.stdout = backup
    output = stream.getvalue()
    stream.close()

    return output


def clean_cmd_filename(cmd, fignum, output_dir):
    if ' -f ' not in cmd and ' --filename ' not in cmd:
        figname = 'fig_' + str(fignum) + '.png'
        filename = os.path.join(output_dir, figname)
        cmd += ' --filename ' + filename
    else:
        raise ValueError
    return cmd, figname


def exec_and_cap_cmd(cmd, fignum, shell, mdstr, output_dir):
    plotting_cmd = is_plotting_command(cmd)
    if plotting_cmd:
        cmd, figfile = clean_cmd_filename(cmd, fignum, output_dir)
        fignum += 1
    # do the comand and grab the output, and close the code markup
    output = do_and_cap(shell, cmd)
    output = '\n'.join(['    ' + s for s in output.split('\n')])
    mdstr += '\n' + output
    if plotting_cmd:
        mdstr += "\n![{}]({})\n".format(cmd, figfile)
    cmd = ''
    return cmd, mdstr, fignum


def mdread(f, output_dir, shell):
    """Reads a .bql file and converts it to markdown.

    Captures text and figure output and saves all code and image assets to
    a directory. Using `mdread` requires some special care on behalf of the
    user. See `writing a BQL script`.

    Parameters
    ----------
    f : file
        The .bql file to convert.
    output_dir : string
        The name of an output directory where all assets will be saved.
    shell : bayeslite.Shell
        The shell object from which the function was called.

    Returns
    -------
    mdstr : string
        A string of markdown

    Notes
    -----
    When writing a script for mdread, do not use ``--filename`` arguments for
    plotting commands. `mdread` will generate filenames that ensure assets are
    saved alongside the markdown.
    """
    if not isinstance(f, file):
        raise TypeError('f should be a file.')
    lines = f.read().split('\n')

    cont = False
    cmd = ''
    mdstr = ''
    last_type = None
    fignum = 0

    # XXX: The first three chracters are stripped from comments, because it is
    # assumed that a space immediately follows '--'. I should probably change
    # this in future.
    line = lines[0]
    last_type = get_line_type(line)
    if last_type == 'code':
        mdstr += "\n"
        cmd = line.strip()
        mdstr += '    ' + line
    elif last_type == 'comment':
        mdstr += line[3:].rstrip() + '\n'
    else:
        mdstr += line

    for i, line in enumerate(lines[1:]):
        linetype = get_line_type(line)
        cont = is_continuation(line)
        if cont and linetype != last_type:
            raise ValueError
        if cont:
            if linetype == 'code':
                cmd += ' ' + line.strip()
                mdstr += '\n' + '    ' + CPROMPT + line
        else:
            if last_type == 'code':
                mdstr += '\n'
                cmd, mdstr, fignum = exec_and_cap_cmd(cmd, fignum, shell,
                                                      mdstr, output_dir)
            if linetype == 'comment':
                mdstr += line[3:].rstrip() + '\n'
            elif linetype == 'code':
                cmd = line
                mdstr += '\n    ' + BPROMPT + line
                if i == len(lines)-1:
                    last_cmd, mdstr, fignum = exec_and_cap_cmd(
                        cmd, fignum, shell, mdstr, output_dir)
            else:
                mdstr += '\n'
        last_type = linetype

    if last_type == 'code' and cont:
        last_cmd, mdstr, fignum = exec_and_cap_cmd(cmd, fignum, shell, mdstr,
                                                   output_dir)

    with open(os.path.join(output_dir, 'README.md'), 'w') as f:
        f.write(mdstr)

    return mdstr


def unicorn():
    """It's not a unicorn at all!"""
    unicorn = """
                                 ,-""   `.
                               ,'  _   e )`-._
                              /  ,' `-._<.===-'
                             /  /
                            /  ;
                _          /   ;
   (`._    _.-"" ""--..__,'    |
   <_  `-""                     \\
    <`-                          :
     (__   <__.                  ;
       `-.   '-.__.      _.'    /
          \\      `-.__,-'    _,'
           `._    ,    /__,-'
              ""._\\__,'< <____
                   | |  `----.`.
                   | |        \\ `.
                   ; |___      \\-``
                   \\   --<
                    `.`.<
               hjw    `-'
    """
    return unicorn
